package org.apache.bcel.classfile;

import java.util.Arrays;
import java.util.Iterator;
import java.io.DataOutputStream;
import org.apache.bcel.Const;
import java.io.IOException;
import java.io.DataInput;

public class ConstantPool implements Cloneable, Node, Iterable<Constant>
{
    private Constant[] constantPool;
    
    private static String escape(final String str) {
        final int len = str.length();
        final StringBuilder buf = new StringBuilder(len + 5);
        final char[] ch = str.toCharArray();
        for (int i = 0; i < len; ++i) {
            switch (ch[i]) {
                case '\n': {
                    buf.append("\\n");
                    break;
                }
                case '\r': {
                    buf.append("\\r");
                    break;
                }
                case '\t': {
                    buf.append("\\t");
                    break;
                }
                case '\b': {
                    buf.append("\\b");
                    break;
                }
                case '\"': {
                    buf.append("\\\"");
                    break;
                }
                default: {
                    buf.append(ch[i]);
                    break;
                }
            }
        }
        return buf.toString();
    }
    
    public ConstantPool(final Constant[] constantPool) {
        this.constantPool = constantPool;
    }
    
    public ConstantPool(final DataInput input) throws IOException {
        final int constantPoolCount = input.readUnsignedShort();
        this.constantPool = new Constant[constantPoolCount];
        for (int i = 1; i < constantPoolCount; ++i) {
            this.constantPool[i] = Constant.readConstant(input);
            final byte tag = this.constantPool[i].getTag();
            if (tag == 6 || tag == 5) {
                ++i;
            }
        }
    }
    
    @Override
    public void accept(final Visitor v) {
        v.visitConstantPool(this);
    }
    
    public String constantToString(Constant c) throws IllegalArgumentException {
        final byte tag = c.getTag();
        String str = null;
        switch (tag) {
            case 7: {
                final int i = ((ConstantClass)c).getNameIndex();
                c = this.getConstantUtf8(i);
                str = Utility.compactClassName(((ConstantUtf8)c).getBytes(), false);
                break;
            }
            case 8: {
                final int i = ((ConstantString)c).getStringIndex();
                c = this.getConstantUtf8(i);
                str = "\"" + escape(((ConstantUtf8)c).getBytes()) + "\"";
                break;
            }
            case 1: {
                str = ((ConstantUtf8)c).getBytes();
                break;
            }
            case 6: {
                str = String.valueOf(((ConstantDouble)c).getBytes());
                break;
            }
            case 4: {
                str = String.valueOf(((ConstantFloat)c).getBytes());
                break;
            }
            case 5: {
                str = String.valueOf(((ConstantLong)c).getBytes());
                break;
            }
            case 3: {
                str = String.valueOf(((ConstantInteger)c).getBytes());
                break;
            }
            case 12: {
                str = this.constantToString(((ConstantNameAndType)c).getNameIndex(), (byte)1) + " " + this.constantToString(((ConstantNameAndType)c).getSignatureIndex(), (byte)1);
                break;
            }
            case 9:
            case 10:
            case 11: {
                str = this.constantToString(((ConstantCP)c).getClassIndex(), (byte)7) + "." + this.constantToString(((ConstantCP)c).getNameAndTypeIndex(), (byte)12);
                break;
            }
            case 15: {
                final ConstantMethodHandle cmh = (ConstantMethodHandle)c;
                str = Const.getMethodHandleName(cmh.getReferenceKind()) + " " + this.constantToString(cmh.getReferenceIndex(), this.getConstant(cmh.getReferenceIndex()).getTag());
                break;
            }
            case 16: {
                final ConstantMethodType cmt = (ConstantMethodType)c;
                str = this.constantToString(cmt.getDescriptorIndex(), (byte)1);
                break;
            }
            case 18: {
                final ConstantInvokeDynamic cid = (ConstantInvokeDynamic)c;
                str = cid.getBootstrapMethodAttrIndex() + ":" + this.constantToString(cid.getNameAndTypeIndex(), (byte)12);
                break;
            }
            case 17: {
                final ConstantDynamic cd = (ConstantDynamic)c;
                str = cd.getBootstrapMethodAttrIndex() + ":" + this.constantToString(cd.getNameAndTypeIndex(), (byte)12);
                break;
            }
            case 19: {
                final int i = ((ConstantModule)c).getNameIndex();
                c = this.getConstantUtf8(i);
                str = Utility.compactClassName(((ConstantUtf8)c).getBytes(), false);
                break;
            }
            case 20: {
                final int i = ((ConstantPackage)c).getNameIndex();
                c = this.getConstantUtf8(i);
                str = Utility.compactClassName(((ConstantUtf8)c).getBytes(), false);
                break;
            }
            default: {
                throw new IllegalArgumentException("Unknown constant type " + tag);
            }
        }
        return str;
    }
    
    public String constantToString(final int index, final byte tag) {
        return this.constantToString(this.getConstant(index, tag));
    }
    
    public ConstantPool copy() {
        ConstantPool c = null;
        try {
            c = (ConstantPool)this.clone();
            c.constantPool = new Constant[this.constantPool.length];
            for (int i = 1; i < this.constantPool.length; ++i) {
                if (this.constantPool[i] != null) {
                    c.constantPool[i] = this.constantPool[i].copy();
                }
            }
        }
        catch (final CloneNotSupportedException ex) {}
        return c;
    }
    
    public void dump(final DataOutputStream file) throws IOException {
        final int size = Math.min(this.constantPool.length, 65535);
        file.writeShort(size);
        for (int i = 1; i < size; ++i) {
            if (this.constantPool[i] != null) {
                this.constantPool[i].dump(file);
            }
        }
    }
    
    public <T extends Constant> T getConstant(final int index) throws ClassFormatException {
        return this.getConstant(index, (Class<T>)Constant.class);
    }
    
    public <T extends Constant> T getConstant(final int index, final byte tag) throws ClassFormatException {
        return this.getConstant(index, tag, (Class<T>)Constant.class);
    }
    
    public <T extends Constant> T getConstant(final int index, final byte tag, final Class<T> castTo) throws ClassFormatException {
        final T c = this.getConstant(index);
        if (c.getTag() != tag) {
            throw new ClassFormatException("Expected class `" + Const.getConstantName(tag) + "' at index " + index + " and got " + c);
        }
        return c;
    }
    
    public <T extends Constant> T getConstant(final int index, final Class<T> castTo) throws ClassFormatException {
        if (index >= this.constantPool.length || index < 0) {
            throw new ClassFormatException("Invalid constant pool reference: " + index + ". Constant pool size is: " + this.constantPool.length);
        }
        final T c = castTo.cast(this.constantPool[index]);
        if (c == null && index != 0) {
            final Constant prev = this.constantPool[index - 1];
            if (prev == null || (prev.getTag() != 6 && prev.getTag() != 5)) {
                throw new ClassFormatException("Constant pool at index " + index + " is null.");
            }
        }
        return c;
    }
    
    public ConstantInteger getConstantInteger(final int index) {
        return this.getConstant(index, (byte)3, ConstantInteger.class);
    }
    
    public Constant[] getConstantPool() {
        return this.constantPool;
    }
    
    public String getConstantString(final int index, final byte tag) throws IllegalArgumentException {
        int i = 0;
        switch (tag) {
            case 7: {
                i = this.getConstant(index, ConstantClass.class).getNameIndex();
                break;
            }
            case 8: {
                i = this.getConstant(index, ConstantString.class).getStringIndex();
                break;
            }
            case 19: {
                i = this.getConstant(index, ConstantModule.class).getNameIndex();
                break;
            }
            case 20: {
                i = this.getConstant(index, ConstantPackage.class).getNameIndex();
                break;
            }
            case 1: {
                return this.getConstantUtf8(index).getBytes();
            }
            default: {
                throw new IllegalArgumentException("getConstantString called with illegal tag " + tag);
            }
        }
        return this.getConstantUtf8(i).getBytes();
    }
    
    public ConstantUtf8 getConstantUtf8(final int index) throws ClassFormatException {
        return this.getConstant(index, (byte)1, ConstantUtf8.class);
    }
    
    public int getLength() {
        return (this.constantPool == null) ? 0 : this.constantPool.length;
    }
    
    @Override
    public Iterator<Constant> iterator() {
        return Arrays.stream(this.constantPool).iterator();
    }
    
    public void setConstant(final int index, final Constant constant) {
        this.constantPool[index] = constant;
    }
    
    public void setConstantPool(final Constant[] constantPool) {
        this.constantPool = constantPool;
    }
    
    @Override
    public String toString() {
        final StringBuilder buf = new StringBuilder();
        for (int i = 1; i < this.constantPool.length; ++i) {
            buf.append(i).append(")").append(this.constantPool[i]).append("\n");
        }
        return buf.toString();
    }
}
